@@ -104,7 +104,7 @@ struct Er {
 |   The following flag is TRUE if entity reference hash table needs
 |   to be initialized.
 |
-|   Hash table is used read-only, therefore just one copy, protected with
+|   Hash table is used read-only, therefore, just one copy, protected with
 |   mutex when used in threading environments. The mutex is used only for
 |   initial setup of the table.
 |
@@ -174,7 +174,7 @@ static Er er_sequences[] = {
 \---------------------------------------------------------------------------*/
 static void ErInit (void)
 {
-    size_t i;  /* For looping thru the list of entity references */
+    size_t i;  /* For looping through the list of entity references */
     int h;  /* The hash on a entity */
 
     for(i=0; i<sizeof(er_sequences)/sizeof(er_sequences[0]); i++){
@@ -335,9 +335,10 @@ XML_SimpleParse (
     char        *xml,   /* XML string  */
     int         *pos,   /* Index of next unparsed character in xml */
     domDocument *doc,
-    domNode     *parent_nodeOld,
+    domNode     *parent,
     int          ignoreWhiteSpaces,
     int          keepCDATA,
+    int          forest,
     char       **errStr
 ) {
     register int   c;          /* Next character of the input file */
@@ -346,7 +347,7 @@ XML_SimpleParse (
     int            saved;
     int            hasContent;
     domNode       *node;
-    domNode       *parent_node = NULL;
+    domNode       *parent_node = parent;
     domTextNode   *tnode;
     domAttrNode   *attrnode, *lastAttr, *attrList;
     int            ampersandSeen = 0;
@@ -1028,6 +1029,10 @@ XML_SimpleParse (
             }
         }
     }
+    if (forest && parent_node == parent) {
+        FREE ((char *) activeNS);
+        return TCL_OK;
+    }
     RetError("Unexpected end",(x - xml) );
 
 } /* XML_SimpleParse */
@@ -1046,19 +1051,42 @@ XML_SimpleParseDocument (
     char    *xml,              /* Complete text of the file being parsed  */
     int      ignoreWhiteSpaces,
     int      keepCDATA,
+    int      forest,
     char    *baseURI,
     Tcl_Obj *extResolver,
     int     *pos,
     char   **errStr
 ) {
     domDocument   *doc = domCreateDoc(baseURI, 0);
-
+    domNode *node = NULL;
+    Tcl_HashEntry *h;
+    int hnew;
+    
     if (extResolver) {
         doc->extResolver = tdomstrdup (Tcl_GetString (extResolver));
     }
+
+    if (forest) {
+        // Create umbreall tag
+        h = Tcl_CreateHashEntry(&HASHTAB(doc,tdom_tagNames), "forestroot",
+                                &hnew);
+        node = (domNode*) domAlloc(sizeof(domNode));
+        memset(node, 0, sizeof(domNode));
+        node->nodeType      = ELEMENT_NODE;
+        node->nodeName      = (char *)&(h->key);
+        node->ownerDocument = doc;
+        doc->rootNode->firstChild = node;
+        doc->rootNode->lastChild = node;
+    }
     
     *pos = 0;
-    XML_SimpleParse (xml, pos, doc, NULL, ignoreWhiteSpaces, keepCDATA, errStr);
+    XML_SimpleParse (xml, pos, doc, node, ignoreWhiteSpaces, keepCDATA,
+                     forest, errStr);
+    if (forest) {
+        doc->rootNode->firstChild = node->firstChild;
+        doc->rootNode->lastChild = node->lastChild;
+        domFree ((void*)node);
+    }
     domSetDocumentElement (doc);
 
     return doc;
